from typing import Callable, Optional, Union
from websockets.sync.client import connect
import pydantic
import websockets
from lmnr.types import (
    DeregisterDebuggerRequest,
    NodeFunction,
    NodeInput,
    RegisterDebuggerRequest,
    SDKError,
    ToolCallError,
    ToolCallRequest,
    ToolCallResponse,
)
import uuid
import json
from threading import Thread


class RemoteDebugger:
    def __init__(
        self,
        project_api_key: str,
        tools: Union[dict[str, NodeFunction], list[Callable[..., NodeInput]]] = [],
    ):
        # for simplicity and backwards compatibility, we allow the user to pass a list
        if isinstance(tools, list):
            tools = {f.__name__: NodeFunction(f.__name__, f) for f in tools}

        self.project_api_key = project_api_key
        self.url = "wss://api.lmnr.ai/v2/endpoint/ws"
        self.tools = tools
        self.thread = Thread(target=self._run)
        self.stop_flag = False
        self.session = None

    def start(self) -> Optional[str]:
        self.stop_flag = False
        self.session = self._generate_session_id()
        self.thread.start()
        return self.session

    def stop(self):
        self.stop_flag = True
        self.thread.join()
        self.session = None
        # python allows running threads only once, so we need to create
        # a new thread
        # in case the user wants to start the debugger again
        self.thread = Thread(target=self._run)

    def _run(self):
        assert self.session is not None, "Session ID not set"
        request = RegisterDebuggerRequest(debuggerSessionId=self.session)
        with connect(
            self.url,
            additional_headers={"Authorization": f"Bearer {self.project_api_key}"},
        ) as websocket:
            websocket.send(request.model_dump_json())
            print(self._format_session_id_and_registerd_functions())
            req_id = None

            while not self.stop_flag:
                try:
                    # blocks the thread until a message
                    # is received or a timeout (3 seconds) occurs
                    message = websocket.recv(3)
                except TimeoutError:
                    continue
                except websockets.exceptions.ConnectionClosedError:
                    print("Connection closed. Please restart the debugger.")
                    return
                try:
                    tool_call = ToolCallRequest.model_validate_json(message)
                    req_id = tool_call.reqId
                except Exception:
                    raise SDKError(f"Invalid message received:\n{message}")
                matching_tool = self.tools.get(tool_call.toolCall.function.name)
                if matching_tool is None:
                    error_message = (
                        f"Tool {tool_call.toolCall.function.name} not found"
                        + ". Registered tools: "
                        + ", ".join(self.tools.keys())
                    )
                    e = ToolCallError(error=error_message, reqId=req_id)
                    websocket.send(e.model_dump_json())
                    continue
                tool = matching_tool.function

                # default the arguments to an empty dictionary
                arguments = {}
                try:
                    arguments = json.loads(tool_call.toolCall.function.arguments)
                except Exception:
                    pass
                try:
                    response = tool(**arguments)
                except Exception as e:
                    error_message = (
                        "Error occurred while running tool" + f"{tool.__name__}: {e}"
                    )
                    e = ToolCallError(error=error_message, reqId=req_id)
                    websocket.send(e.model_dump_json())
                    continue
                formatted_response = None
                try:
                    formatted_response = ToolCallResponse(
                        reqId=tool_call.reqId, response=response
                    )
                except pydantic.ValidationError:
                    formatted_response = ToolCallResponse(
                        reqId=tool_call.reqId, response=str(response)
                    )
                websocket.send(formatted_response.model_dump_json())
            websocket.send(
                DeregisterDebuggerRequest(
                    debuggerSessionId=self.session, deregister=True
                ).model_dump_json()
            )

    def _generate_session_id(self) -> str:
        return uuid.uuid4().urn[9:]

    def _format_session_id_and_registerd_functions(self) -> str:
        registered_functions = ",\n".join(["- " + k for k in self.tools.keys()])
        return f"""
========================================
Debugger Session ID:
{self.session}
========================================

Registered functions:
{registered_functions}

========================================
"""



from .sdk.endpoint import Laminar
from .types import ChatMessage, EndpointRunError, EndpointRunResponse, NodeInput
from .sdk.remote_debugger import RemoteDebugger as LaminarRemoteDebugger
from .sdk.registry import Registry as Pipeline



import os
from pathlib import Path
import zipfile


def zip_directory(directory_path: Path, zip_file_path: Path):
    with zipfile.ZipFile(zip_file_path, "w", zipfile.ZIP_DEFLATED) as zipf:
        for root, _, files in os.walk(directory_path):
            for file in files:
                file_path = os.path.join(root, file)
                arcname = os.path.relpath(file_path, directory_path)
                zipf.write(file_path, arcname)



from typing import Callable

from lmnr.types import NodeFunction, NodeInput


class Registry:
    """
    Class to register and resolve node functions based on their node names.

    Node names cannot have space in their name.
    """

    functions: dict[str, NodeFunction]

    def __init__(self):
        self.functions = {}

    def add(self, node_name: str, function: Callable[..., NodeInput]):
        self.functions[node_name] = NodeFunction(node_name, function)

    def func(self, node_name: str):
        def decorator(f: Callable[..., NodeInput]):
            self.add(node_name, f)
            return f

        return decorator

    def get(self, node_name: str) -> Callable[..., NodeInput]:
        return self.functions[node_name].function



from dataclasses import dataclass
from typing import Optional
import uuid

from lmnr.cli.parser.nodes import Handle, NodeFunctions
from lmnr.types import NodeInput


@dataclass
class InputNode(NodeFunctions):
    id: uuid.UUID
    name: str
    outputs: list[Handle]
    input: Optional[NodeInput]

    def handles_mapping(
        self, output_handle_id_to_node_name: dict[str, str]
    ) -> list[tuple[str, str]]:
        return []

    def node_type(self) -> str:
        return "Input"

    def config(self) -> dict:
        return {}



from dataclasses import dataclass

import uuid

from lmnr.cli.parser.nodes import Handle, NodeFunctions
from lmnr.cli.parser.utils import map_handles


@dataclass
class Dataset:
    id: uuid.UUID
    # created_at: datetime
    # project_id: uuid.UUID
    # name: str
    # indexed_on: Optional[str]

    @classmethod
    def from_dict(cls, dataset_dict: dict) -> "Dataset":
        return cls(
            id=uuid.UUID(dataset_dict["id"]),
        )


@dataclass
class SemanticSearchNode(NodeFunctions):
    id: uuid.UUID
    name: str
    inputs: list[Handle]
    outputs: list[Handle]
    inputs_mappings: dict[uuid.UUID, uuid.UUID]
    limit: int
    threshold: float
    template: str
    datasets: list[Dataset]

    def handles_mapping(
        self, output_handle_id_to_node_name: dict[str, str]
    ) -> list[tuple[str, str]]:
        return map_handles(
            self.inputs, self.inputs_mappings, output_handle_id_to_node_name
        )

    def node_type(self) -> str:
        return "SemanticSearch"

    def config(self) -> dict:
        return {
            "limit": self.limit,
            "threshold": self.threshold,
            "template": self.template,
            "datasource_ids": [str(dataset.id) for dataset in self.datasets],
            "datasource_ids_list": str([str(dataset.id) for dataset in self.datasets]),
        }



from dataclasses import dataclass
import uuid

from lmnr.cli.parser.nodes import Handle, NodeFunctions
from lmnr.cli.parser.utils import map_handles


@dataclass
class OutputNode(NodeFunctions):
    id: uuid.UUID
    name: str
    inputs: list[Handle]
    outputs: list[Handle]
    inputs_mappings: dict[uuid.UUID, uuid.UUID]

    def handles_mapping(
        self, output_handle_id_to_node_name: dict[str, str]
    ) -> list[tuple[str, str]]:
        return map_handles(
            self.inputs, self.inputs_mappings, output_handle_id_to_node_name
        )

    def node_type(self) -> str:
        return "Output"

    def config(self) -> dict:
        return {}



from typing import Any, Union
import uuid

from lmnr.cli.parser.nodes import Handle
from lmnr.cli.parser.nodes.code import CodeNode
from lmnr.cli.parser.nodes.condition import ConditionNode
from lmnr.cli.parser.nodes.input import InputNode
from lmnr.cli.parser.nodes.json_extractor import JsonExtractorNode
from lmnr.cli.parser.nodes.llm import LLMNode
from lmnr.cli.parser.nodes.output import OutputNode
from lmnr.cli.parser.nodes.router import Route, RouterNode
from lmnr.cli.parser.nodes.semantic_search import (
    Dataset,
    SemanticSearchNode,
)
from lmnr.types import NodeInput, ChatMessage


def node_input_from_json(json_val: Any) -> NodeInput:
    if isinstance(json_val, str):
        return json_val
    elif isinstance(json_val, list):
        return [ChatMessage.model_validate(msg) for msg in json_val]
    else:
        raise ValueError(f"Invalid NodeInput value: {json_val}")


Node = Union[
    InputNode,
    OutputNode,
    ConditionNode,
    LLMNode,
    RouterNode,
    SemanticSearchNode,
    CodeNode,
    JsonExtractorNode,
]


def node_from_dict(node_dict: dict) -> Node:
    if node_dict["type"] == "Input":
        return InputNode(
            id=uuid.UUID(node_dict["id"]),
            name=node_dict["name"],
            outputs=[Handle.from_dict(handle) for handle in node_dict["outputs"]],
            input=node_input_from_json(node_dict["input"]),
        )
    elif node_dict["type"] == "Output":
        return OutputNode(
            id=uuid.UUID(node_dict["id"]),
            name=node_dict["name"],
            inputs=[Handle.from_dict(handle) for handle in node_dict["inputs"]],
            outputs=[Handle.from_dict(handle) for handle in node_dict["outputs"]],
            inputs_mappings={
                uuid.UUID(k): uuid.UUID(v)
                for k, v in node_dict["inputsMappings"].items()
            },
        )
    elif node_dict["type"] == "Condition":
        return ConditionNode(
            id=uuid.UUID(node_dict["id"]),
            name=node_dict["name"],
            inputs=[Handle.from_dict(handle) for handle in node_dict["inputs"]],
            outputs=[Handle.from_dict(handle) for handle in node_dict["outputs"]],
            inputs_mappings={
                uuid.UUID(k): uuid.UUID(v)
                for k, v in node_dict["inputsMappings"].items()
            },
            condition=node_dict["condition"],
        )
    elif node_dict["type"] == "LLM":
        return LLMNode(
            id=uuid.UUID(node_dict["id"]),
            name=node_dict["name"],
            inputs=[Handle.from_dict(handle) for handle in node_dict["inputs"]],
            dynamic_inputs=[
                Handle.from_dict(handle) for handle in node_dict["dynamicInputs"]
            ],
            outputs=[Handle.from_dict(handle) for handle in node_dict["outputs"]],
            inputs_mappings={
                uuid.UUID(k): uuid.UUID(v)
                for k, v in node_dict["inputsMappings"].items()
            },
            prompt=node_dict["prompt"],
            model=node_dict["model"],
            model_params=(
                node_dict["modelParams"] if "modelParams" in node_dict else None
            ),
            stream=False,
            structured_output_enabled=node_dict.get("structuredOutputEnabled", False),
            structured_output_max_retries=node_dict.get(
                "structuredOutputMaxRetries", 0
            ),
            structured_output_schema=node_dict.get("structuredOutputSchema", None),
            structured_output_schema_target=node_dict.get(
                "structuredOutputSchemaTarget", None
            ),
        )
    elif node_dict["type"] == "Router":
        return RouterNode(
            id=uuid.UUID(node_dict["id"]),
            name=node_dict["name"],
            inputs=[Handle.from_dict(handle) for handle in node_dict["inputs"]],
            outputs=[Handle.from_dict(handle) for handle in node_dict["outputs"]],
            inputs_mappings={
                uuid.UUID(k): uuid.UUID(v)
                for k, v in node_dict["inputsMappings"].items()
            },
            routes=[Route(name=route["name"]) for route in node_dict["routes"]],
            has_default_route=node_dict["hasDefaultRoute"],
        )
    elif node_dict["type"] == "SemanticSearch":
        return SemanticSearchNode(
            id=uuid.UUID(node_dict["id"]),
            name=node_dict["name"],
            inputs=[Handle.from_dict(handle) for handle in node_dict["inputs"]],
            outputs=[Handle.from_dict(handle) for handle in node_dict["outputs"]],
            inputs_mappings={
                uuid.UUID(k): uuid.UUID(v)
                for k, v in node_dict["inputsMappings"].items()
            },
            limit=node_dict["limit"],
            threshold=node_dict["threshold"],
            template=node_dict["template"],
            datasets=[Dataset.from_dict(ds) for ds in node_dict["datasets"]],
        )
    elif node_dict["type"] == "Code":
        return CodeNode(
            id=uuid.UUID(node_dict["id"]),
            name=node_dict["name"],
            inputs=[Handle.from_dict(handle) for handle in node_dict["inputs"]],
            outputs=[Handle.from_dict(handle) for handle in node_dict["outputs"]],
            inputs_mappings={
                uuid.UUID(k): uuid.UUID(v)
                for k, v in node_dict["inputsMappings"].items()
            },
            code=node_dict["code"],
            fn_name=node_dict["fnName"],
        )
    elif node_dict["type"] == "JsonExtractor":
        return JsonExtractorNode(
            id=uuid.UUID(node_dict["id"]),
            name=node_dict["name"],
            inputs=[Handle.from_dict(handle) for handle in node_dict["inputs"]],
            outputs=[Handle.from_dict(handle) for handle in node_dict["outputs"]],
            inputs_mappings={
                uuid.UUID(k): uuid.UUID(v)
                for k, v in node_dict["inputsMappings"].items()
            },
            template=node_dict["template"],
        )
    else:
        raise ValueError(f"Node type {node_dict['type']} not supported")



import json
from pydantic.alias_generators import to_snake
import pydantic
import requests
from lmnr.types import (
    EndpointRunError, EndpointRunResponse, NodeInput, EndpointRunRequest,
    ToolCallError, ToolCallRequest, ToolCallResponse, SDKError
)
from typing import Callable, Optional
from websockets.sync.client import connect

class Laminar:
    project_api_key: Optional[str] = None
    def __init__(self, project_api_key: str):
        """Initialize the Laminar object with your project API key

        Args:
            project_api_key (str):
                Project api key. Generate or view your keys
                in the project settings in the Laminar dashboard.
        """        
        self.project_api_key = project_api_key
        self.url = 'https://api.lmnr.ai/v2/endpoint/run'
        self.ws_url = 'wss://api.lmnr.ai/v2/endpoint/ws'
    
    def run (
        self,
        endpoint: str,
        inputs: dict[str, NodeInput],
        env: dict[str, str] = {},
        metadata: dict[str, str] = {},
        tools: list[Callable[..., NodeInput]] = [],
    ) -> EndpointRunResponse:
        """Runs the endpoint with the given inputs

        Args:
            endpoint (str): name of the Laminar endpoint
            inputs (dict[str, NodeInput]):
                inputs to the endpoint's target pipeline.
                Keys in the dictionary must match input node names
            env (dict[str, str], optional):
                Environment variables for the pipeline execution.
                Defaults to {}.
            metadata (dict[str, str], optional):
                any custom metadata to be stored
                with execution trace. Defaults to {}.
            tools (list[Callable[..., NodeInput]], optional):
                List of callable functions the execution can call as tools.
                If specified and non-empty, a bidirectional communication
                with Laminar API through websocket will be established.
                Defaults to [].

        Returns:
            EndpointRunResponse: response object containing the outputs
        
        Raises:
            ValueError: if project API key is not set
            EndpointRunError: if the endpoint run fails
            SDKError: if an error occurs on client side during the execution
        """        
        if self.project_api_key is None:
            raise ValueError(
                'Please initialize the Laminar object with' 
                ' your project API key'
            )
        if tools:
            return self._run_websocket(endpoint, inputs, env, metadata, tools)
        return self._run(endpoint, inputs, env, metadata)
    
    def _run(
        self,
        endpoint: str,
        inputs: dict[str, NodeInput],
        env: dict[str, str] = {},
        metadata: dict[str, str] = {}
    ) -> EndpointRunResponse:
        try:
            request = EndpointRunRequest(
                inputs = inputs,
                endpoint = endpoint,
                env = env,
                metadata = metadata
            )
        except Exception as e:
            raise ValueError(f'Invalid request: {e}')
        response = requests.post(
            self.url, 
            json=json.loads(request.model_dump_json()),
            headers={'Authorization': f'Bearer {self.project_api_key}'}
        )
        if response.status_code != 200:
            raise EndpointRunError(response)
        try:
            resp_json = response.json()
            keys = list(resp_json.keys())
            for key in keys:
                value = resp_json[key]
                del resp_json[key]
                resp_json[to_snake(key)] = value
            return EndpointRunResponse(**resp_json)
        except:
            raise EndpointRunError(response)

    def _run_websocket(
        self,
        endpoint: str,
        inputs: dict[str, NodeInput],
        env: dict[str, str] = {},
        metadata: dict[str, str] = {},
        tools: list[Callable[..., NodeInput]] = [],
    ) -> EndpointRunResponse:
        try:
            request = EndpointRunRequest(
                inputs = inputs,
                endpoint = endpoint,
                env = env,
                metadata = metadata
            )
        except Exception as e:
            raise ValueError(f'Invalid request: {e}')

        with connect(
            self.ws_url,
            additional_headers={
                'Authorization': f'Bearer {self.project_api_key}'
            }
        ) as websocket:
            websocket.send(request.model_dump_json())
            req_id = None

            while True:
                message = websocket.recv()
                try:
                    tool_call = ToolCallRequest.model_validate_json(message)
                    req_id = tool_call.req_id
                    matching_tools = [
                        tool for tool in tools
                        if tool.__name__ == tool_call.toolCall.function.name
                    ]
                    if not matching_tools:
                        raise SDKError(
                            f'Tool {tool_call.toolCall.function.name} not found.'
                            ' Registered tools: '
                            f'{", ".join([tool.__name__ for tool in tools])}'
                        )
                    tool = matching_tools[0]
                        # default the arguments to an empty dictionary
                    if tool.__name__ == tool_call.toolCall.function.name:
                        arguments = {}
                        try:
                            arguments = json.loads(tool_call.toolCall.function.arguments)
                        except:
                            pass
                        try:
                            response = tool(**arguments)
                        except Exception as e:
                            error_message = 'Error occurred while running tool' +\
                             f'{tool.__name__}: {e}'
                            e = ToolCallError(error=error_message, reqId=req_id)
                            websocket.send(e.model_dump_json())
                        formatted_response = None
                        try:
                            formatted_response = ToolCallResponse(
                                reqId=tool_call.reqId,
                                response=response
                            )
                        except pydantic.ValidationError as e:
                            formatted_response = ToolCallResponse(
                                reqId=tool_call.reqId,
                                response=str(response)
                            )
                        websocket.send(
                            formatted_response.model_dump_json()
                        )
                except pydantic.ValidationError as e:
                    message_json = json.loads(message)
                    keys = list(message_json.keys())
                    for key in keys:
                        value = message_json[key]
                        del message_json[key]
                        message_json[to_snake(key)] = value
                    result = EndpointRunResponse.model_validate(message_json)
                    websocket.close()
                    return result
                except Exception as e:
                    raise SDKError(f'Error communicating to backend through websocket {e}')



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import Optional
import uuid


@dataclass
class Handle:
    id: uuid.UUID
    name: Optional[str]

    @classmethod
    def from_dict(cls, dict: dict) -> "Handle":
        return cls(
            id=uuid.UUID(dict["id"]),
            name=(dict["name"] if "name" in dict else None),
        )


@abstractmethod
class NodeFunctions(metaclass=ABCMeta):
    @abstractmethod
    def handles_mapping(
        self, output_handle_id_to_node_name: dict[str, str]
    ) -> list[tuple[str, str]]:
        """
        Returns a list of tuples mapping from this node's input
        handle name to the unique name of the previous node.

        Assumes previous node has only one output.
        """
        pass

    @abstractmethod
    def node_type(self) -> str:
        pass

    @abstractmethod
    def config(self) -> dict:
        """
        Returns a dictionary of node-specific configuration.

        E.g. prompt and model name for LLM node.
        """
        pass



from lmnr.cli.parser.utils import replace_spaces_with_underscores
from .nodes.types import node_from_dict


def runnable_graph_to_template_vars(graph: dict) -> dict:
    """
    Convert a runnable graph to template vars to be rendered in a cookiecutter context.
    """
    node_id_to_node_name = {}
    output_handle_id_to_node_name: dict[str, str] = {}
    for node in graph["nodes"].values():
        # override node names in the graph itself to be safe
        node["name"] = replace_spaces_with_underscores(node["name"])

        node_id_to_node_name[node["id"]] = node["name"]
        for handle in node["outputs"]:
            output_handle_id_to_node_name[handle["id"]] = node["name"]

    tasks = []
    for node_obj in graph["nodes"].values():
        node = node_from_dict(node_obj)
        handles_mapping = node.handles_mapping(output_handle_id_to_node_name)
        node_type = node.node_type()

        unique_handles = set([handle_name for (handle_name, _) in handles_mapping])

        tasks.append(
            {
                "name": node.name,
                "function_name": f"run_{node.name}",
                "node_type": node_type,
                "handles_mapping": handles_mapping,
                # since we map from to to from, all 'to's won't repeat
                "input_handle_names": [
                    handle_name for (handle_name, _) in handles_mapping
                ],
                "handle_args": ", ".join(
                    [f"{handle_name}: NodeInput" for handle_name in unique_handles]
                ),
                "prev": [],
                "next": [],
                "config": node.config(),
            }
        )

    for to, from_ in graph["pred"].items():
        # TODO: Make "tasks" a hashmap from node id (as str!) to task
        to_task = [task for task in tasks if task["name"] == node_id_to_node_name[to]][
            0
        ]
        from_tasks = []
        for f in from_:
            from_tasks.append(
                [task for task in tasks if task["name"] == node_id_to_node_name[f]][0]
            )

        for from_task in from_tasks:
            to_task["prev"].append(from_task["name"])
            from_task["next"].append(node_id_to_node_name[to])

    # Return as a hashmap due to cookiecutter limitations, investigate later.
    return {task["name"]: task for task in tasks}



import requests
import pydantic
import uuid
from typing import Callable, Union, Optional


class ChatMessage(pydantic.BaseModel):
    role: str
    content: str


class ConditionedValue(pydantic.BaseModel):
    condition: str
    value: "NodeInput"


NodeInput = Union[str, list[ChatMessage], ConditionedValue]  # TypeAlias


class EndpointRunRequest(pydantic.BaseModel):
    inputs: dict[str, NodeInput]
    endpoint: str
    env: dict[str, str] = pydantic.Field(default_factory=dict)
    metadata: dict[str, str] = pydantic.Field(default_factory=dict)


class EndpointRunResponse(pydantic.BaseModel):
    outputs: dict[str, dict[str, NodeInput]]
    run_id: str


class EndpointRunError(Exception):
    error_code: str
    error_message: str

    def __init__(self, response: requests.Response):
        try:
            resp_json = response.json()
            self.error_code = resp_json["error_code"]
            self.error_message = resp_json["error_message"]
            super().__init__(self.error_message)
        except Exception:
            super().__init__(response.text)

    def __str__(self) -> str:
        try:
            return str(
                {"error_code": self.error_code, "error_message": self.error_message}
            )
        except Exception:
            return super().__str__()


class SDKError(Exception):
    def __init__(self, error_message: str):
        super().__init__(error_message)


class ToolCallFunction(pydantic.BaseModel):
    name: str
    arguments: str


class ToolCall(pydantic.BaseModel):
    id: Optional[str]
    type: Optional[str]
    function: ToolCallFunction


# TODO: allow snake_case and manually convert to camelCase
class ToolCallRequest(pydantic.BaseModel):
    reqId: uuid.UUID
    toolCall: ToolCall


class ToolCallResponse(pydantic.BaseModel):
    reqId: uuid.UUID
    response: NodeInput


class ToolCallError(pydantic.BaseModel):
    reqId: uuid.UUID
    error: str


class RegisterDebuggerRequest(pydantic.BaseModel):
    debuggerSessionId: str


class DeregisterDebuggerRequest(pydantic.BaseModel):
    debuggerSessionId: str
    deregister: bool


class NodeFunction:
    node_name: str
    function: Callable[..., NodeInput]

    def __init__(self, node_name: str, function: Callable[..., NodeInput]):
        self.node_name = node_name
        self.function = function



from dataclasses import dataclass
from typing import Optional
import uuid

from lmnr.cli.parser.nodes import Handle, NodeFunctions
from lmnr.cli.parser.utils import map_handles


@dataclass
class LLMNode(NodeFunctions):
    id: uuid.UUID
    name: str
    inputs: list[Handle]
    dynamic_inputs: list[Handle]
    outputs: list[Handle]
    inputs_mappings: dict[uuid.UUID, uuid.UUID]
    prompt: str
    model: str
    model_params: Optional[str]
    stream: bool
    structured_output_enabled: bool
    structured_output_max_retries: int
    structured_output_schema: Optional[str]
    structured_output_schema_target: Optional[str]

    def handles_mapping(
        self, output_handle_id_to_node_name: dict[str, str]
    ) -> list[tuple[str, str]]:
        combined_inputs = self.inputs + self.dynamic_inputs
        return map_handles(
            combined_inputs, self.inputs_mappings, output_handle_id_to_node_name
        )

    def node_type(self) -> str:
        return "LLM"

    def config(self) -> dict:
        # For easier access in the template separate the provider and model here
        provider, model = self.model.split(":", maxsplit=1)

        return {
            "prompt": self.prompt,
            "provider": provider,
            "model": model,
            "model_params": self.model_params,
            "stream": self.stream,
            "enable_structured_output": self.structured_output_enabled
            and self.structured_output_schema is not None,
            "structured_output_max_retries": self.structured_output_max_retries,
            "structured_output_schema": self.structured_output_schema,
            "structured_output_schema_target_str": (
                "None"
                if self.structured_output_schema_target is None
                else f'"{self.structured_output_schema_target}"'
            ),
        }
